@@ -1,16 +1,17 @@
 /*
- * Copyright (c) 2014 Hong Xu <hong AT topbug DOT net>
+ * Copyright (c) 2014-2019 Hong Xu <hong AT topbug DOT net>
+ * Copyright (c) 2018 Sven Strickroth <email AT cs-ware DOT de>
  * All rights reserved.
- * 
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
- * 
+ *
  * 1. Redistributions of source code must retain the above copyright notice,
  *    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright notice,
  *    this list of conditions and the following disclaimer in the documentation
  *    and/or other materials provided with the distribution.
- * 
+ *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
@@ -28,14 +29,18 @@
 
 #include <ctype.h>
 #include <string.h>
-#include <pcre.h>
+#include <pcre2.h>
 
-#define oom() { return -1; }
+#define utarray_oom() { return -2; }
 #include "utarray.h"
 #include "misc.h"
+#include "util.h"
 
 #include "ec_glob.h"
 
+/* Special characters */
+const char ec_special_chars[] = "?[]\\*-{},";
+
 typedef struct int_pair
 {
     int     num1;
@@ -52,9 +57,10 @@ static const UT_icd ut_int_pair_icd = {sizeof(int_pair),NULL,NULL,NULL};
     p += string_len; \
 } while(0)
 
-#define PATTERN_MAX  300
+#define PATTERN_MAX  4097
 /*
- * Whether the string matches the given glob pattern
+ * Whether the string matches the given glob pattern. Return 0 if successful, return -1 if a PCRE
+ * error or other regex error occurs, and return -2 if an OOM outside PCRE occurs.
  */
 EDITORCONFIG_LOCAL
 int ec_glob(const char *pattern, const char *string)
@@ -67,15 +73,14 @@ int ec_glob(const char *pattern, const char *string)
     char *                    pcre_str_end;
     int                       brace_level = 0;
     _Bool                     is_in_bracket = 0;
-    const char *              error_msg;
-    int                       erroffset;
-    pcre *                    re;
+    int                       error_code;
+    size_t                    erroffset;
+    pcre2_code *              re;
     int                       rc;
-    int *                     pcre_result;
-    size_t                    pcre_result_len;
+    size_t *                  pcre_result;
+    pcre2_match_data *        pcre_match_data;
     char                      l_pattern[2 * PATTERN_MAX];
-    int                       pattern_length = strlen(pattern);
-    _Bool                     are_brace_paired;
+    _Bool                     are_braces_paired = 1;
     UT_array *                nums;     /* number ranges */
     int                       ret = 0;
 
@@ -83,6 +88,7 @@ int ec_glob(const char *pattern, const char *string)
     p_pcre = pcre_str + 1;
     pcre_str_end = pcre_str + 2 * PATTERN_MAX;
 
+    /* Determine whether curly braces are paired */
     {
         int     left_count = 0;
         int     right_count = 0;
@@ -96,16 +102,23 @@ int ec_glob(const char *pattern, const char *string)
 
             if (*c == '}')
                 ++ right_count;
-            if (*c == '{')
+            else if (*c == '{')
                 ++ left_count;
+
+            if (right_count > left_count)
+            {
+                are_braces_paired = 0;
+                break;
+            }
         }
 
-        are_brace_paired = (right_count == left_count);
+        if (right_count != left_count)
+            are_braces_paired = 0;
     }
 
     /* used to search for {num1..num2} case */
-    re = pcre_compile("^\\{[\\+\\-]?\\d+\\.\\.[\\+\\-]?\\d+\\}$", 0,
-            &error_msg, &erroffset, NULL);
+    re = pcre2_compile("^\\{[\\+\\-]?\\d+\\.\\.[\\+\\-]?\\d+\\}$", PCRE2_ZERO_TERMINATED, 0,
+            &error_code, &erroffset, NULL);
     if (!re)        /* failed to compile */
         return -1;
 
@@ -126,7 +139,7 @@ int ec_glob(const char *pattern, const char *string)
 
             break;
         case '?':
-            *(p_pcre ++) = '.';
+            STRING_CAT(p_pcre, "[^/]", pcre_str_end);
             break;
         case '*':
             if (*(c+1) == '*')      /* case of ** */
@@ -169,11 +182,21 @@ int ec_glob(const char *pattern, const char *string)
                 {
                     char *           right_bracket = strchr(c, ']');
 
+                    if (!right_bracket)  /* The right bracket may not exist */
+                        right_bracket = c + strlen(c);
+
                     strcat(p_pcre, "\\");
                     strncat(p_pcre, c, right_bracket - c);
-                    strcat(p_pcre, "\\]");
+                    if (*right_bracket)  /* right_bracket is a bracket */
+                        strcat(p_pcre, "\\]");
                     p_pcre += strlen(p_pcre);
                     c = right_bracket;
+                    if (!*c)
+                        /* end of string, meaning that right_bracket is not a
+                         * bracket. Then we go back one character to make the
+                         * parsing end normally for the counter in the "for"
+                         * loop. */
+                        c -= 1;
                     break;
                 }
             }
@@ -202,7 +225,7 @@ int ec_glob(const char *pattern, const char *string)
 
             break;
         case '{':
-            if (!are_brace_paired)
+            if (!are_braces_paired)
             {
                 STRING_CAT(p_pcre, "\\{", pcre_str_end);
                 break;
@@ -235,11 +258,13 @@ int ec_glob(const char *pattern, const char *string)
                 {
                     const char *        double_dots;
                     int_pair            pair;
-                    int                 pcre_res[3];
+
+                    pcre2_match_data *  match_data = pcre2_match_data_create_from_pattern(re, NULL);
 
                     /* Check the case of {num1..num2} */
-                    rc = pcre_exec(re, NULL, c, (int) (cc - c + 1), 0, 0,
-                            pcre_res, 3);
+                    rc = pcre2_match(re, c, cc - c + 1, 0, 0, match_data, NULL);
+
+                    pcre2_match_data_free(match_data);
 
                     if (rc < 0)    /* not {num1..num2} case */
                     {
@@ -253,8 +278,8 @@ int ec_glob(const char *pattern, const char *string)
 
                     /* Get the range */
                     double_dots = strstr(c, "..");
-                    pair.num1 = atoi(c + 1);
-                    pair.num2 = atoi(double_dots + 2);
+                    pair.num1 = ec_atoi(c + 1);
+                    pair.num2 = ec_atoi(double_dots + 2);
 
                     utarray_push_back(nums, &pair);
 
@@ -270,7 +295,7 @@ int ec_glob(const char *pattern, const char *string)
             break;
 
         case '}':
-            if (!are_brace_paired)
+            if (!are_braces_paired)
             {
                 STRING_CAT(p_pcre, "\\}", pcre_str_end);
                 break;
@@ -309,34 +334,31 @@ int ec_glob(const char *pattern, const char *string)
 
     *(p_pcre ++) = '$';
 
-    pcre_free(re); /* ^\\d+\\.\\.\\d+$ */
+    pcre2_code_free(re); /* ^\\d+\\.\\.\\d+$ */
 
-    re = pcre_compile(pcre_str, 0, &error_msg, &erroffset, NULL);
+    re = pcre2_compile(pcre_str, PCRE2_ZERO_TERMINATED, 0, &error_code, &erroffset, NULL);
 
     if (!re)        /* failed to compile */
+    {
+        utarray_free(nums);
         return -1;
+    }
 
-    pcre_result_len = 3 * (utarray_len(nums) + 1);
-    pcre_result = (int *) calloc(pcre_result_len, sizeof(int_pair));
-    rc = pcre_exec(re, NULL, string, (int) strlen(string), 0, 0,
-            pcre_result, pcre_result_len);
+    pcre_match_data = pcre2_match_data_create_from_pattern(re, NULL);
+    rc = pcre2_match(re, string, strlen(string), 0, 0, pcre_match_data, NULL);
 
     if (rc < 0)     /* failed to match */
     {
-        int         ret;
-        if (rc == PCRE_ERROR_NOMATCH)
+        if (rc == PCRE2_ERROR_NOMATCH)
             ret = EC_GLOB_NOMATCH;
         else
             ret = rc;
 
-        pcre_free(re);
-        free(pcre_result);
-        utarray_free(nums);
-
-        return ret;
+        goto cleanup;
     }
 
     /* Whether the numbers are in the desired range? */
+    pcre_result = pcre2_get_ovector_pointer(pcre_match_data);
     for(p = (int_pair *) utarray_front(nums), i = 1; p;
             ++ i, p = (int_pair *) utarray_next(nums, p))
     {
@@ -350,7 +372,11 @@ int ec_glob(const char *pattern, const char *string)
             break;
 
         num_string = strndup(substring_start, substring_length);
-        num = atoi(num_string);
+        if (num_string == NULL) {
+          ret = -2;
+          goto cleanup;
+        }
+        num = ec_atoi(num_string);
         free(num_string);
 
         if (num < p->num1 || num > p->num2) /* not matched */
@@ -360,8 +386,10 @@ int ec_glob(const char *pattern, const char *string)
     if (p != NULL)      /* numbers not matched */
         ret = EC_GLOB_NOMATCH;
 
-    pcre_free(re);
-    free(pcre_result);
+ cleanup:
+
+    pcre2_code_free(re);
+    pcre2_match_data_free(pcre_match_data);
     utarray_free(nums);
 
     return ret;
